import asyncio
from enum import Enum
from typing import Literal, List, Type, DefaultDict
from collections import UserDict, defaultdict

from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage

# from edsl.enums import pricing, TokenPricing
from edsl.enums import get_token_pricing
from edsl.jobs.tasks.task_status_enum import TaskStatus

InterviewTokenUsageMapping = DefaultDict[str, InterviewTokenUsage]

from edsl.jobs.interviews.InterviewStatistic import InterviewStatistic
from edsl.jobs.interviews.InterviewStatisticsCollection import (
    InterviewStatisticsCollection,
)


class JobsRunnerStatusData:
    # pricing = pricing

    def status_dict(self, interviews):
        status = []
        for interview in interviews:
            # model = interview.model
            status.append(interview.interview_status)

        return status
        # return model_to_status

    def status_counts(self, interviews):
        model_to_status = defaultdict(InterviewStatusDictionary)

        for interview in interviews:
            model = interview.model
            model_to_status[model] += interview.interview_status

        # breakpoint()
        return model_to_status.values()

    def generate_status_summary(
        self,
        completed_tasks: List[Type[asyncio.Task]],
        elapsed_time: float,
        interviews: List[Type["Interview"]],
    ) -> InterviewStatisticsCollection:
        """Generate a summary of the status of the job runner.

        :param completed_tasks: list of completed tasks
        :param elapsed_time: time elapsed since the start of the job
        :param interviews: list of interviews to be conducted
        """

        models_to_tokens = defaultdict(InterviewTokenUsage)
        model_to_status = defaultdict(InterviewStatusDictionary)

        waiting_dict = defaultdict(int)

        interview_statistics = InterviewStatisticsCollection()

        for interview in interviews:
            model = interview.model
            models_to_tokens[model] += interview.token_usage
            model_to_status[model] += interview.interview_status
            waiting_dict[model] += interview.interview_status.waiting

        interview_statistics.add_stat(
            InterviewStatistic(
                "elapsed_time", value=elapsed_time, digits=1, units="sec."
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "total_interviews_requested", value=len(interviews), units=""
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "completed_interviews", value=len(completed_tasks), units=""
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "percent_complete",
                value=len(completed_tasks) / len(interviews) * 100
                if len(interviews) > 0
                else "NA",
                digits=0,
                units="%",
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "average_time_per_interview",
                value=elapsed_time / len(completed_tasks) if completed_tasks else "NA",
                digits=1,
                units="sec.",
            )
        )
        interview_statistics.add_stat(
            InterviewStatistic(
                "task_remaining", value=len(interviews) - len(completed_tasks), units=""
            )
        )
        number_remaining = len(interviews) - len(completed_tasks)
        time_per_task = (
            elapsed_time / len(completed_tasks) if len(completed_tasks) > 0 else "NA"
        )
        estimated_time_remaining = (
            number_remaining * time_per_task if time_per_task != "NA" else "NA"
        )

        interview_statistics.add_stat(
            InterviewStatistic(
                "estimated_time_remaining",
                value=estimated_time_remaining,
                digits=1,
                units="sec.",
            )
        )
        model_queues_info = []
        for model, num_waiting in waiting_dict.items():
            model_info = self._get_model_info(model, num_waiting, models_to_tokens)
            model_queues_info.append(model_info)

        interview_statistics["model_queues"] = model_queues_info

        return interview_statistics

    def _get_model_info(
        self,
        model: str,
        num_waiting: int,
        models_to_tokens: InterviewTokenUsageMapping,
    ):
        """Get the status of a model."""
        # if model.model not in self.pricing:
        #     #raise ValueError(f"Model {model.model} not found in pricing")
        #     import warning
        #     warning.warn(f"Model {model.model} not found in pricing")

        prices = get_token_pricing(model.model)

        model_info = {
            "model_name": model.model,
            "TPM_limit_k": model.TPM / 1000,
            "RPM_limit_k": model.RPM / 1000,
            "num_tasks_waiting": num_waiting,
            "token_usage_info": [],
        }

        token_usage_types = ["new_token_usage", "cached_token_usage"]
        for token_usage_type in token_usage_types:
            cache_info = self._get_token_usage_info(
                token_usage_type, models_to_tokens, model, prices
            )
            model_info["token_usage_info"].append(cache_info)

        return model_info

    def _get_token_usage_info(
        self,
        cache_status: Literal["new_token_usage", "cached_token_usage"],
        models_to_tokens: InterviewTokenUsageMapping,
        model: str,
        prices: "TokenPricing",
    ):
        cache_info = {"cache_status": cache_status, "details": []}
        token_usage = getattr(models_to_tokens[model], cache_status)
        for token_type in ["prompt_tokens", "completion_tokens"]:
            tokens = getattr(token_usage, token_type)
            cache_info["details"].append(
                {
                    "type": token_type,
                    "tokens": tokens,
                }
            )
        cache_info["cost"] = f"${token_usage.cost(prices):.5f}"
        return cache_info



class QuestionErrors(Exception):
    pass


class QuestionCreationValidationError(QuestionErrors):
    pass


class QuestionResponseValidationError(QuestionErrors):
    pass


class QuestionAnswerValidationError(QuestionErrors):
    pass


class QuestionAttributeMissing(QuestionErrors):
    pass


class QuestionSerializationError(QuestionErrors):
    pass


class QuestionScenarioRenderError(QuestionErrors):
    pass


class QuestionMissingTypeError(QuestionErrors):
    pass


class QuestionBadTypeError(QuestionErrors):
    pass



"""Budget question instructions."""
import textwrap

from edsl.prompts.QuestionInstructionsBase import QuestionInstuctionsBase


class Budget(QuestionInstuctionsBase):
    """Budget question instructions."""

    question_type = "budget"
    model = "gpt-4-1106-preview"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are 
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        Return a valid JSON formatted as follows, with a dictionary for your "answer"
        where the keys are the option numbers and the values are the amounts you want 
        to allocate to the options, and the sum of the values is {{budget_sum}}:
        {"answer": {<put dict of option numbers and allocation amounts here>},
        "comment": "<put explanation here>"}
        Example response for a budget of 100 and 4 options: 
        {"answer": {"0": 25, "1": 25, "2": 25, "3": 25},
        "comment": "I allocated 25 to each option."}
        There must be an allocation listed for each item (including 0).
        """
    )



from typing import Dict, Any

from edsl.prompts.Prompt import Prompt
from edsl.utilities.decorators import sync_wrapper, jupyter_nb_handler
from edsl.prompts.registry import get_classes as prompt_lookup
from edsl.exceptions import QuestionScenarioRenderError


class PromptConstructorMixin:
    def construct_system_prompt(self) -> Prompt:
        """Construct the system prompt for the LLM call."""

        agent_instructions = self._get_agent_instructions_prompt()
        persona_prompt = self._get_persona_prompt()

        return (
            agent_instructions
            + " " * int(len(persona_prompt.text) > 0)
            + persona_prompt
        )

    def _get_persona_prompt(self) -> Prompt:
        """Get the persona prompt.

        The is the description of the agent to the LLM.

        The agent_persona is constructed when the Agent is created.
        If the agent is passed a template for "agent_trait_presentation_template" that is used to construct the persona.
        If it does not exist, the persona is looked up in the prompt registry
        """
        if not hasattr(self.agent, "agent_persona"):
            applicable_prompts = prompt_lookup(
                component_type="agent_persona",
                model=self.model.model,
            )
            persona_prompt_template = applicable_prompts[0]()
        else:
            persona_prompt_template = self.agent.agent_persona

        # TODO: This multiple passing of agent traits - not sure if it is necessary. Not harmful.
        if undefined := persona_prompt_template.undefined_template_variables(
            self.agent.traits
            | {"traits": self.agent.traits}
            | {"codebook": self.agent.codebook}
            | {"traits": self.agent.traits}
        ):
            raise QuestionScenarioRenderError(
                f"Agent persona still has variables that were not rendered: {undefined}"
            )

        persona_prompt = persona_prompt_template.render(
            self.agent.traits | {"traits": self.agent.traits},
            codebook=self.agent.codebook,
            traits=self.agent.traits,
        )

        if persona_prompt.has_variables:
            raise QuestionScenarioRenderError(
                "Agent persona still has variables that were not rendered."
            )
        return persona_prompt

    def _get_agent_instructions_prompt(self) -> Prompt:
        """Get the agent instructions prompt."""
        applicable_prompts = prompt_lookup(
            component_type="agent_instructions",
            model=self.model.model,
        )
        if len(applicable_prompts) == 0:
            raise Exception("No applicable prompts found")
        return applicable_prompts[0](text=self.agent.instruction)

    def _get_question_instructions(self) -> Prompt:
        """Get the instructions for the question."""
        # applicable_prompts = prompt_lookup(
        #    component_type="question_instructions",
        #    question_type=self.question.question_type,
        #    model=self.model.model,
        # )
        ## Get the question instructions and renders with the scenario & question.data
        # question_prompt = applicable_prompts[0]()
        question_prompt = self.question.get_instructions(model=self.model.model)

        undefined_template_variables = question_prompt.undefined_template_variables(
            self.question.data | self.scenario
        )
        if undefined_template_variables:
            print(undefined_template_variables)
            raise QuestionScenarioRenderError(
                "Question instructions still has variables."
            )

        return question_prompt.render(self.question.data | self.scenario)

    def construct_user_prompt(self) -> Prompt:
        """Construct the user prompt for the LLM call."""
        user_prompt = self._get_question_instructions()
        if self.memory_plan is not None:
            user_prompt += self.create_memory_prompt(
                self.question.question_name
            ).render(self.scenario)
        return user_prompt

    def get_prompts(self) -> Dict[str, Prompt]:
        """Get both prompts for the LLM call."""
        system_prompt = self.construct_system_prompt()
        user_prompt = self.construct_user_prompt()
        return {
            "user_prompt": user_prompt,
            "system_prompt": system_prompt,
        }


if __name__ == "__main__":
    from edsl import Model
    from edsl import Agent

    a = Agent(
        instruction="You are a happy-go lucky agent.",
        traits={"feeling": "happy", "age": "Young at heart"},
        codebook={"feeling": "Feelings right now", "age": "Age in years"},
        trait_presentation_template="",
    )
    p = PromptConstructorMixin()
    p.model = Model(Model.available()[0])
    p.agent = a
    instructions = p._get_agent_instructions_prompt()
    repr(instructions)

    persona = p._get_persona_prompt()
    repr(persona)



from __future__ import annotations
from typing import Optional, Union
import random

from jinja2 import Template

from edsl.utilities import random_string
from edsl.questions.descriptors import QuestionOptionsDescriptor
from edsl.questions.QuestionBase import QuestionBase
from edsl.scenarios import Scenario


class QuestionMultipleChoice(QuestionBase):
    """This question prompts the agent to select one option from a list of options."""

    question_type = "multiple_choice"
    purpose = "When options are known and limited"
    question_options: list[str] = QuestionOptionsDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        question_options: list[str],
    ):
        """Instantiate a new QuestionMultipleChoice.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param question_options: The options the agent should select from.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionMultipleChoice.default_instructions`.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.question_options = question_options

    ################
    # Answer methods
    ################
    def _validate_answer(
        self, answer: dict[str, Union[str, int]]
    ) -> dict[str, Union[str, int]]:
        """Validate the answer."""
        self._validate_answer_template_basic(answer)
        self._validate_answer_multiple_choice(answer)
        return answer

    def _translate_answer_code_to_answer(self, answer_code, scenario: Scenario = None):
        """Translate the answer code to the actual answer."""
        scenario = scenario or Scenario()
        translated_options = [
            Template(str(option)).render(scenario) for option in self.question_options
        ]
        return translated_options[int(answer_code)]

    def _simulate_answer(
        self, human_readable: bool = True
    ) -> dict[str, Union[int, str]]:
        """Simulate a valid answer for debugging purposes."""
        if human_readable:
            answer = random.choice(self.question_options)
        else:
            answer = random.choice(range(len(self.question_options)))
        return {
            "answer": answer,
            "comment": random_string(),
        }

    ################
    # Example
    ################
    @classmethod
    def example(cls) -> QuestionMultipleChoice:
        """Return an example instance."""
        return cls(
            question_text="How are you?",
            question_options=["Good", "Great", "OK", "Bad"],
            question_name="how_feeling",
        )


def main():
    """Create an example QuestionMultipleChoice and test its methods."""
    from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice

    q = QuestionMultipleChoice.example()
    q.question_text
    q.question_options
    q.question_name
    # validate an answer
    q._validate_answer({"answer": 0, "comment": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer(0, {})
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import asyncio
import aiohttp
import json
from typing import Any

from edsl import CONFIG

from edsl.language_models.LanguageModel import LanguageModel


def replicate_model_factory(model_name, base_url, api_token):
    class ReplicateLanguageModelBase(LanguageModel):
        _model_ = (
            model_name  # Example model name, replace with actual model name if needed
        )
        _parameters_ = {
            "temperature": 0.1,
            "topK": 50,
            "topP": 0.9,
            "max_new_tokens": 500,
            "min_new_tokens": -1,
            "repetition_penalty": 1.15,
            #       "version": "5fe0a3d7ac2852264a25279d1dfb798acbc4d49711d126646594e212cb821749",
            "use_cache": True,
        }
        _api_token = api_token
        _base_url = base_url

        async def async_execute_model_call(
            self, user_prompt: str, system_prompt: str = ""
        ) -> dict[str, Any]:
            self.api_token = self._api_token
            self.headers = {
                "Authorization": f"Token {self.api_token}",
                "Content-Type": "application/json",
            }
            # combined_prompt = f"{system_prompt} {user_prompt}".strip()
            # print(f"Prompt: {combined_prompt}")
            data = {
                #          "version": self._parameters_["version"],
                "input": {
                    "debug": False,
                    "top_k": self._parameters_["topK"],
                    "top_p": self._parameters_["topP"],
                    "prompt": user_prompt,
                    "system_prompt": system_prompt,
                    "temperature": self._parameters_["temperature"],
                    "max_new_tokens": self._parameters_["max_new_tokens"],
                    "min_new_tokens": self._parameters_["min_new_tokens"],
                    "prompt_template": "{prompt}",
                    "repetition_penalty": self._parameters_["repetition_penalty"],
                },
            }

            async with aiohttp.ClientSession() as session:
                async with session.post(
                    self._base_url, headers=self.headers, data=json.dumps(data)
                ) as response:
                    raw_response_text = await response.text()
                    data = json.loads(raw_response_text)
                    print(f"This was the data returned by the model:{data}")
                    prediction_url = data["urls"]["get"]

                while True:
                    async with session.get(
                        prediction_url, headers=self.headers
                    ) as get_response:
                        if get_response.status != 200:
                            # Handle non-success status codes appropriately
                            return None

                        get_data = await get_response.text()
                        get_data = json.loads(get_data)
                        if get_data["status"] == "succeeded":
                            return get_data
                        await asyncio.sleep(1)

        def parse_response(self, raw_response: dict[str, Any]) -> str:
            data = "".join(raw_response["output"])
            print(f"This is what the model returned: {data}")
            return data

    return ReplicateLanguageModelBase



import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
ROOT_DIR = os.path.dirname(BASE_DIR)

from edsl.__version__ import __version__
from edsl.config import Config, CONFIG
from edsl.agents.Agent import Agent
from edsl.agents.AgentList import AgentList
from edsl.questions import (
    QuestionFreeText,
    QuestionMultipleChoice,
    QuestionNumerical,
    QuestionCheckBox,
)
from edsl.scenarios.Scenario import Scenario
from edsl.scenarios.ScenarioList import ScenarioList
from edsl.utilities.interface import print_dict_with_rich
from edsl.surveys.Survey import Survey
from edsl.language_models.registry import Model
from edsl.questions.question_registry import Question
from edsl.results.Results import Results
from edsl.data.Cache import Cache
from edsl.coop.coop import Coop



"""A mixin for visualizing the flow of a survey."""
import pydot
import tempfile
from IPython.display import Image
from edsl.utilities import is_notebook
from edsl.surveys.base import RulePriority, EndOfSurvey


class SurveyFlowVisualizationMixin:
    """A mixin for visualizing the flow of a survey."""

    def show_flow(self):
        """Create an image showing the flow of users through the survey."""
        # Create a graph object
        graph = pydot.Dot(graph_type="digraph")

        # Add nodes
        for index, question in enumerate(self.questions):
            graph.add_node(
                pydot.Node(
                    f"Q{index}", label=f"{question.question_name}", shape="ellipse"
                )
            )

        # Add an "EndOfSurvey" node
        graph.add_node(
            pydot.Node("EndOfSurvey", label="End of Survey", shape="rectangle")
        )

        # Add edges for normal flow
        num_questions = len(self.questions)
        for index in range(num_questions - 1):  # From Q1 to Q3
            graph.add_edge(pydot.Edge(f"Q{index}", f"Q{index+1}"))

        graph.add_edge(pydot.Edge(f"Q{num_questions-1}", "EndOfSurvey"))

        relevant_rules = [
            rule
            for rule in self.rule_collection
            if rule.priority > RulePriority.DEFAULT.value
        ]

        colors = [
            "blue",
            "red",
            "purple",
            "orange",
            "pink",
            "brown",
            "grey",
            "cyan",
            "green",
        ]
        rule_colors = {
            rule: colors[i % len(colors)] for i, rule in enumerate(relevant_rules)
        }

        for rule in relevant_rules:
            color = rule_colors[rule]
            edge_label = f"if {rule.expression}"
            source_node = f"Q{rule.current_q}"
            target_node = (
                f"Q{rule.next_q}" if rule.next_q != EndOfSurvey else "EndOfSurvey"
            )
            edge = pydot.Edge(
                source_node, target_node, label=edge_label, color=color, fontcolor=color
            )
            graph.add_edge(edge)

        with tempfile.NamedTemporaryFile(delete=False, suffix=".png") as tmp_file:
            try:
                graph.write_png(tmp_file.name)
            except FileNotFoundError:
                print(
                    """File not found. Most likely it's because you don't have graphviz installed. Please install it and try again.
                        It's 
                        $ sudo apt-get install graphviz 
                        on Ubuntu.
                    """
                )

            if is_notebook():
                display(Image(tmp_file.name))
            else:
                import os
                import sys

                if os.name == "nt":  # Windows
                    os.system(f"start {tmp_file.name}")
                elif os.name == "posix":  # macOS, Linux, Unix, etc.
                    os.system(
                        f"open {tmp_file.name}"
                        if sys.platform == "darwin"
                        else f"xdg-open {tmp_file.name}"
                    )



import json
import requests
import threading
from edsl.coop import Coop


class ReportErrors:
    def __init__(self, task_history):
        self.task_history = task_history
        self.email = None

    @property
    def data(self):
        return {
            "text": self.task_history.to_dict(),
            "email": self.email,
        }

    def get_email(self, timeout=10):
        """Helper method to get user input with a timeout."""
        input_queue = []

        def input_thread_method():
            email_input = input(
                "Please enter your email address (if you want us to get in touch): "
            )
            input_queue.append(email_input)

        input_thread = threading.Thread(target=input_thread_method)
        input_thread.start()
        input_thread.join(timeout=timeout)

        if input_queue:
            self.email = input_queue[0]
        else:
            print("No input received within the timeout period.")

    def upload(self):
        coop = Coop()
        coop.send_error_message(error_data=self.data)


def main():
    from edsl.jobs.interviews.ReportErrors import ReportErrors

    class TaskHistory:
        def __init__(self, data):
            self.data = data

        def to_dict(self):
            """Converts the internal data of the task history to a dictionary format."""
            return self.data

    task_history_data = {
        "task": "Example Task",
        "status": "Completed",
        "details": "This is an example of a task history.",
    }
    task_history = TaskHistory(task_history_data)

    reporter = ReportErrors(task_history)
    # one without email
    reporter.upload()
    # one with email
    reporter.email = "fake@gmail.com"
    reporter.upload()



"""Mixin for fetching data from results."""
from functools import partial
from itertools import chain


class ResultsFetchMixin:
    """Mixin for fetching data from results."""

    def _fetch_list(self, data_type, key) -> list:
        """
        Return a list of values from the data for a given data type and key.

        Uses the filtered data, not the original data.

        Example:
        >>> r = Results.create_example()
        >>> r._fetch_list('answer', 'how_feeling')
        ['Bad', 'Bad', 'Great', 'Great']
        """
        returned_list = []
        for row in self.data:
            returned_list.append(row.sub_dicts[data_type].get(key, None))

        return returned_list


if __name__ == "__main__":
    from edsl.results import Results

    r = Results.example()



"""
The Results object is the result of running a survey. 
It is not typically instantiated directly, but is returned by the run method of a `Job` object.
"""
from __future__ import annotations
import json
import io
import sys
from collections import UserList, defaultdict
from typing import Optional
from rich.console import Console

from simpleeval import EvalWithCompoundTypes
from typing import Any, Type, Union

from edsl.exceptions.results import (
    ResultsBadMutationstringError,
    ResultsColumnNotFoundError,
    ResultsInvalidNameError,
    ResultsMutateError,
    ResultsFilterError,
)
from edsl.agents import Agent
from edsl.language_models.LanguageModel import LanguageModel
from edsl.results.Dataset import Dataset
from edsl.results.Result import Result
from edsl.results.ResultsExportMixin import ResultsExportMixin
from edsl.scenarios import Scenario
from edsl.surveys import Survey
from edsl.data.Cache import Cache
from edsl.utilities import (
    is_gzipped,
    is_valid_variable_name,
    print_public_methods_with_doc,
    shorten_string,
    is_notebook,
)

import json
from pygments import highlight
from pygments.lexers import JsonLexer
from pygments.formatters import HtmlFormatter
from IPython.display import HTML


from edsl.results.ResultsDBMixin import ResultsDBMixin


class Mixins(ResultsExportMixin, ResultsDBMixin):
    pass


from edsl.Base import Base
from edsl.results.ResultsFetchMixin import ResultsFetchMixin


class Mixins(ResultsExportMixin, ResultsDBMixin, ResultsFetchMixin):
    pass


class Results(UserList, Mixins, Base):
    """
    This class is a UserList of Result objects.

    It is instantiated with a `Survey` and a list of `Result` objects.
    It can be manipulated in various ways with select, filter, mutate, etc.
    It also has a list of created_columns, which are columns that have been created with `mutate` and are not part of the original data.
    """

    known_data_types = [
        "answer",
        "scenario",
        "agent",
        "model",
        "prompt",
        "raw_model_response",
        "iteration",
    ]

    def __init__(
        self,
        survey: Optional[Survey] = None,
        data: Optional[list[Result]] = None,
        created_columns: Optional[list[str]] = None,
        cache: Optional[Cache] = None,
        job_uuid: Optional[str] = None,
        total_results: Optional[int] = None,
    ):
        """Instantiate a `Results` object with a survey and a list of `Result` objects.

        :param survey: A Survey object.
        :param data: A list of Result objects.
        :param created_columns: A list of strings that are created columns.
        :param job_uuid: A string representing the job UUID.
        :param total_results: An integer representing the total number of results.
        """
        super().__init__(data)
        self.survey = survey
        self.created_columns = created_columns or []
        self._job_uuid = job_uuid
        self._total_results = total_results
        self.cache = cache or Cache()

        if hasattr(self, "_add_output_functions"):
            self._add_output_functions()

    ######################
    # Streaming methods
    ######################

    def code(self):
        raise NotImplementedError

    def __getitem__(self, i):
        if isinstance(i, int):
            if isinstance(i, slice):
                # Return a sliced view of the list
                return self.__class__(survey=self.survey, data=self.data[i])
            else:
                # Return a single item
                return self.data[i]
        else:
            return self.to_dict()[i]

    def _update_results(self) -> None:
        if self._job_uuid and len(self.data) < self._total_results:
            results = [
                Result(
                    agent=Agent.from_dict(json.loads(r.agent)),
                    scenario=Scenario.from_dict(json.loads(r.scenario)),
                    model=LanguageModel.from_dict(json.loads(r.model)),
                    iteration=1,
                    answer=json.loads(r.answer),
                )
                for r in CRUD.read_results(self._job_uuid)
            ]
            self.data = results

    def __repr__(self) -> str:
        from rich import print_json
        import json

        print_json(json.dumps(self.to_dict()["data"]))
        return f"Results(data = {self.data}, survey = {repr(self.survey)}, created_columns = {self.created_columns})"

    def _repr_html_(self) -> str:
        json_str = json.dumps(self.to_dict()["data"], indent=4)
        formatted_json = highlight(
            json_str,
            JsonLexer(),
            HtmlFormatter(style="default", full=True, noclasses=True),
        )
        return HTML(formatted_json).data

    def to_dict(self) -> dict[str, Any]:
        """Convert the Results object to a dictionary.

        The dictionary can be quite large, as it includes all of the data in the Results object.

        Example: Illustrating just the keys of the dictionary.

        >>> r = Results.example()
        >>> r.to_dict().keys()
        dict_keys(['data', 'survey', 'created_columns'])
        """
        return {
            "data": [result.to_dict() for result in self.data],
            "survey": self.survey.to_dict(),
            "created_columns": self.created_columns,
            "cache": Cache() if not hasattr(self, "cache") else self.cache.to_dict(),
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> Results:
        """Convert a dictionary to a Results object.

        :param data: A dictionary representation of a Results object.

        Example:

        >>> r = Results.example()
        >>> d = r.to_dict()
        >>> r2 = Results.from_dict(d)
        >>> r == r2
        True
        """
        results = cls(
            survey=Survey.from_dict(data["survey"]),
            data=[Result.from_dict(r) for r in data["data"]],
            created_columns=data.get("created_columns", None),
            cache=Cache.from_dict(data.get("cache")) if "cache" in data else Cache(),
        )
        return results

    ######################
    ## Convenience methods
    ## & Report methods
    ######################
    @property
    def _key_to_data_type(self) -> dict[str, str]:
        """
        Return a mapping of keys (how_feeling, status, etc.) to strings representing data types.

        Objects such as Agent, Answer, Model, Scenario, etc.
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`
        """
        d = {}
        for result in self.data:
            d.update(result.key_to_data_type)
        for column in self.created_columns:
            d[column] = "answer"
        return d

    @property
    def _data_type_to_keys(self) -> dict[str, str]:
        """
        Return a mapping of strings representing data types (objects such as Agent, Answer, Model, Scenario, etc.) to keys (how_feeling, status, etc.)
        - Uses the key_to_data_type property of the Result class.
        - Includes any columns that the user has created with `mutate`

        Example:

        >>> r = Results.example()
        >>> r._data_type_to_keys
        defaultdict(...
        """
        d: dict = defaultdict(set)
        for result in self.data:
            for key, value in result.key_to_data_type.items():
                d[value] = d[value].union(set({key}))
        for column in self.created_columns:
            d["answer"] = d["answer"].union(set({column}))
        return d

    @property
    def columns(self) -> list[str]:
        """Return a list of all of the columns that are in the Results.

        Example:

        >>> r = Results.example()
        >>> r.columns
        ['agent.agent_name', ...
        """
        column_names = [f"{v}.{k}" for k, v in self._key_to_data_type.items()]
        return sorted(column_names)

    @property
    def answer_keys(self) -> dict[str, str]:
        """Return a mapping of answer keys to question text.

        Example:

        >>> r = Results.example()
        >>> r.answer_keys
        {'how_feeling': 'How are you this {{ period }}?', 'how_feeling_yesterday': 'How were you feeling yesterday {{ period }}?'}
        """
        if not self.survey:
            raise Exception("Survey is not defined so no answer keys are available.")

        answer_keys = self._data_type_to_keys["answer"]
        answer_keys = {k for k in answer_keys if "_comment" not in k}
        questions_text = [
            self.survey.get_question(k).question_text for k in answer_keys
        ]
        short_question_text = [shorten_string(q, 80) for q in questions_text]
        initial_dict = dict(zip(answer_keys, short_question_text))
        sorted_dict = {key: initial_dict[key] for key in sorted(initial_dict)}
        return sorted_dict

    @property
    def agents(self) -> list[Agent]:
        """Return a list of all of the agents in the Results.

        Example:

        >>> r = Results.example()
        >>> r.agents
        [Agent(traits = {'status': 'Joyful'}), Agent(traits = {'status': 'Joyful'}), Agent(traits = {'status': 'Sad'}), Agent(traits = {'status': 'Sad'})]
        """
        return [r.agent for r in self.data]

    @property
    def models(self) -> list[Type[LanguageModel]]:
        """Return a list of all of the models in the Results.

        Example:

        >>> r = Results.example()
        >>> r.models[0]
        Model(model = 'gpt-4-1106-preview', parameters={'temperature': 0.5, 'max_tokens': 1000, 'top_p': 1, 'frequency_penalty': 0, 'presence_penalty': 0, 'logprobs': False, 'top_logprobs': 3})
        """
        return [r.model for r in self.data]

    @property
    def scenarios(self) -> list[Scenario]:
        """Return a list of all of the scenarios in the Results.

        Example:

        >>> r = Results.example()
        >>> r.scenarios
        [{'period': 'morning'}, {'period': 'afternoon'}, {'period': 'morning'}, {'period': 'afternoon'}]
        """
        return [r.scenario for r in self.data]

    @property
    def agent_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the Agent data.

        Example:

        >>> r = Results.example()
        >>> r.agent_keys
        ['agent_name', 'status']
        """
        return sorted(self._data_type_to_keys["agent"])

    @property
    def model_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the LanguageModel data.

        >>> r = Results.example()
        >>> r.model_keys
        ['frequency_penalty', 'logprobs', 'max_tokens', 'model', 'presence_penalty', 'temperature', 'top_logprobs', 'top_p']
        """
        return sorted(self._data_type_to_keys["model"])

    @property
    def scenario_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the Scenario data.

        >>> r = Results.example()
        >>> r.scenario_keys
        ['period']
        """
        return sorted(self._data_type_to_keys["scenario"])

    @property
    def question_names(self) -> list[str]:
        """Return a list of all of the question names.

        Example:

        >>> r = Results.example()
        >>> r.question_names
        ['how_feeling', 'how_feeling_yesterday']
        """
        if self.survey is None:
            return []
        return sorted(list(self.survey.question_names))

    @property
    def all_keys(self) -> list[str]:
        """Return a set of all of the keys that are in the Results.

        Example:

        >>> r = Results.example()
        >>> r.all_keys
        ['agent_name', 'frequency_penalty', 'how_feeling', 'how_feeling_yesterday', 'logprobs', 'max_tokens', 'model', 'period', 'presence_penalty', 'status', 'temperature', 'top_logprobs', 'top_p']
        """
        answer_keys = set(self.answer_keys)
        all_keys = (
            answer_keys.union(self.agent_keys)
            .union(self.scenario_keys)
            .union(self.model_keys)
        )
        return sorted(list(all_keys))

    def relevant_columns(self) -> list[str]:
        """Return all of the columns that are in the Results.

        Example:

        >>> r = Results.example()
        >>> r.relevant_columns()
        ['agent', 'agent_name', 'answer', 'frequency_penalty', 'how_feeling', 'how_feeling_comment', 'how_feeling_question_text', 'how_feeling_raw_model_response', 'how_feeling_system_prompt', 'how_feeling_user_prompt', 'how_feeling_yesterday', 'how_feeling_yesterday_comment', 'how_feeling_yesterday_question_text', 'how_feeling_yesterday_raw_model_response', 'how_feeling_yesterday_system_prompt', 'how_feeling_yesterday_user_prompt', 'iteration', 'logprobs', 'max_tokens', 'model', 'period', 'presence_penalty', 'prompt', 'question_text', 'raw_model_response', 'scenario', 'status', 'temperature', 'top_logprobs', 'top_p']
        """
        return sorted(
            set().union(
                *(observation.combined_dict.keys() for observation in self.data)
            )
        )

    def _parse_column(self, column: str) -> tuple[str, str]:
        """
        Parses a column name into a tuple containing a data type and a key.

        >>> r = Results.example()
        >>> r._parse_column("answer.how_feeling")
        ('answer', 'how_feeling')

        The standard way a column is specified is with a dot-separated string, e.g. _parse_column("agent.status")
        But you can also specify a single key, e.g. "status", in which case it will look up the data type.
        """
        if "." in column:
            data_type, key = column.split(".")
        else:
            try:
                data_type, key = self._key_to_data_type[column], column
            except KeyError:
                raise ResultsColumnNotFoundError(f"Column {column} not found in data")
        return data_type, key

    def first(self) -> Result:
        """Return the first observation in the results.

        Example:

        >>> r = Results.example()
        >>> r.first()
        Result(agent...
        """
        return self.data[0]

    def mutate(
        self, new_var_string: str, functions_dict: Optional[dict] = None
    ) -> Results:
        """
        Creates a value in the Results object as if has been asked as part of the survey.

        :param new_var_string: A string that is a valid Python expression.
        :param functions_dict: A dictionary of functions that can be used in the expression. The keys are the function names and the values are the functions themselves.

        It splits the new_var_string at the "=" and uses simple_eval

        Example:

        >>> r = Results.example()
        >>> r.mutate('how_feeling_x = how_feeling + "x"').select('how_feeling_x')
        [{'answer.how_feeling_x': ...
        """
        # extract the variable name and the expression
        if "=" not in new_var_string:
            raise ResultsBadMutationstringError(
                f"Mutate requires an '=' in the string, but '{new_var_string}' doesn't have one."
            )
        raw_var_name, expression = new_var_string.split("=", 1)
        var_name = raw_var_name.strip()
        if not is_valid_variable_name(var_name):
            raise ResultsInvalidNameError(f"{var_name} is not a valid variable name.")

        # create the evaluator
        functions_dict = functions_dict or {}

        def create_evaluator(result: Result) -> EvalWithCompoundTypes:
            return EvalWithCompoundTypes(
                names=result.combined_dict, functions=functions_dict
            )

        def new_result(old_result: Result, var_name: str) -> Result:
            evaluator = create_evaluator(old_result)
            value = evaluator.eval(expression)
            new_result = old_result.copy()
            new_result["answer"][var_name] = value
            return new_result

        try:
            new_data = [new_result(result, var_name) for result in self.data]
        except Exception as e:
            raise ResultsMutateError(f"Error in mutate. Exception:{e}")

        return Results(
            survey=self.survey,
            data=new_data,
            created_columns=self.created_columns + [var_name],
        )

    def select(self, *columns: Union[str, list[str]]) -> Dataset:
        """
        Select data from the results and format it.

        :param columns: A list of strings, each of which is a column name. The column name can be a single key, e.g. "how_feeling", or a dot-separated string, e.g. "answer.how_feeling".

        Example:

        >>> results = Results.example()
        >>> results.select('how_feeling')
        [{'answer.how_feeling': ...
        """

        if not columns or columns == ("*",) or columns == (None,):
            columns = ("*.*",)

        if isinstance(columns[0], list):
            columns = tuple(columns[0])

        def get_data_types_to_return(parsed_data_type):
            if parsed_data_type == "*":  # they want all of the columns
                return self.known_data_types
            else:
                if parsed_data_type not in self.known_data_types:
                    raise Exception(
                        f"Data type {parsed_data_type} not found in data. Did you mean one of {self.known_data_types}"
                    )
                return [parsed_data_type]

        # we're doing to populate this with the data we want to fetch
        to_fetch = defaultdict(list)

        new_data = []
        items_in_order = []
        # iterate through the passed columns
        for column in columns:
            # a user could pass 'result.how_feeling' or just 'how_feeling'
            parsed_data_type, parsed_key = self._parse_column(column)
            data_types = get_data_types_to_return(parsed_data_type)
            found_once = False  # we need to track this to make sure we found the key at least once

            for data_type in data_types:
                # the keys for that data_type e.g.,# if data_type is 'answer', then the keys are 'how_feeling', 'how_feeling_comment', etc.
                relevant_keys = self._data_type_to_keys[data_type]

                for key in relevant_keys:
                    if key == parsed_key or parsed_key == "*":
                        found_once = True
                        to_fetch[data_type].append(key)
                        items_in_order.append(data_type + "." + key)

            if not found_once:
                raise Exception(f"Key {parsed_key} not found in data.")

        for data_type in to_fetch:
            for key in to_fetch[data_type]:
                entries = self._fetch_list(data_type, key)
                new_data.append({data_type + "." + key: entries})

        def sort_by_key_order(dictionary):
            # Extract the single key from the dictionary
            single_key = next(iter(dictionary))
            # Return the index of this key in the list_of_keys
            return items_in_order.index(single_key)

        sorted(new_data, key=sort_by_key_order)

        return Dataset(new_data)

    def sort_by(self, column, reverse: bool = False) -> Results:
        """Sort the results by a column.

        :param column: A string that is a column name.
        :param reverse: A boolean that determines whether to sort in reverse order.

        The column name can be a single key, e.g. "how_feeling", or a dot-separated string, e.g. "answer.how_feeling".

        Example:

        >>> r = Results.example()
        >>> r.sort_by('how_feeling', reverse = False).select('how_feeling').print()
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ Great        │
        ├──────────────┤
        │ OK           │
        ├──────────────┤
        │ OK           │
        ├──────────────┤
        │ Terrible     │
        └──────────────┘
        >>> r.sort_by('how_feeling', reverse = True).select('how_feeling').print()
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ Terrible     │
        ├──────────────┤
        │ OK           │
        ├──────────────┤
        │ OK           │
        ├──────────────┤
        │ Great        │
        └──────────────┘
        """
        data_type, key = self._parse_column(column)

        def to_numeric_if_possible(v):
            try:
                return float(v)
            except:
                return v

        new_data = sorted(
            self.data,
            key=lambda x: to_numeric_if_possible(x.get_value(data_type, key)),
            reverse=reverse,
        )
        return Results(survey=self.survey, data=new_data, created_columns=None)

    def filter(self, expression: str) -> Results:
        """
        Filter based on the given expression and returns the filtered `Results`.

        :param expression: A string expression that evaluates to a boolean. The expression is applied to each element in `Results` to determine whether it should be included in the filtered results.

        The `expression` parameter is a string that must resolve to a boolean value when evaluated against each element in `Results`.
        This expression is used to determine which elements to include in the returned `Results`.

        Example usage: Create an example `Results` instance and apply filters to it:

        >>> r = Results.example()
        >>> r.filter("how_feeling == 'Great'").select('how_feeling').print()
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ Great        │
        └──────────────┘

        Example usage: Using an OR operator in the filter expression.

        >>> r.filter("how_feeling == 'Great' or how_feeling == 'Terrible'").select('how_feeling').print()
        ┏━━━━━━━━━━━━━━┓
        ┃ answer       ┃
        ┃ .how_feeling ┃
        ┡━━━━━━━━━━━━━━┩
        │ Great        │
        ├──────────────┤
        │ Terrible     │
        └──────────────┘
        """

        def create_evaluator(result):
            """Create an evaluator for the given result.
            The 'combined_dict' is a mapping of all values for that Result object.
            """
            return EvalWithCompoundTypes(names=result.combined_dict)

        try:
            # iterates through all the results and evaluates the expression
            new_data = [
                result
                for result in self.data
                if create_evaluator(result).eval(expression)
            ]
        except Exception as e:
            print(f"Exception:{e}")
            raise ResultsFilterError(f"Error in filter. Exception:{e}")

        return Results(survey=self.survey, data=new_data, created_columns=None)

    @classmethod
    def example(cls, debug: bool = False) -> Results:
        """Return an example `Results` object.

        Example usage:

        >>> r = Results.example()

        :param debug: if False, uses actual API calls
        """
        from edsl.jobs import Jobs
        from edsl.data.Cache import Cache

        c = Cache()
        job = Jobs.example()
        results = job.run(cache=c, debug=debug)
        return results

    def rich_print(self):
        """Display an object as a table."""
        pass
        # with io.StringIO() as buf:
        #     console = Console(file=buf, record=True)

        #     for index, result in enumerate(self):
        #         console.print(f"Result {index}")
        #         console.print(result.rich_print())

        #     return console.export_text()

    def __str__(self):
        data = self.to_dict()["data"]
        return json.dumps(data, indent=4)


def main():  # pragma: no cover
    """Call the OpenAI API credits."""
    from edsl.results.Results import Results

    results = Results.example(debug=True)
    print(results.filter("how_feeling == 'Great'").select("how_feeling"))
    print(results.mutate("how_feeling_x = how_feeling + 'x'").select("how_feeling_x"))


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from edsl.inference_services.InferenceServiceABC import InferenceServiceABC


class InferenceServicesCollection:
    def __init__(self, services: list[InferenceServiceABC] = None):
        self.services = services or []

    def available(self):
        total_models = []
        for service in self.services:
            try:
                service_models = service.available()
            except Exception as e:
                print(f"Error getting models for {service._inference_service_}: {e}")
                service_models = []
                continue
            for model in service_models:
                total_models.append([model, service._inference_service_, -1])
        sorted_models = sorted(total_models)
        for i, model in enumerate(sorted_models):
            model[2] = i
            model = tuple(model)
        return sorted_models

    def register(self, service):
        self.services.append(service)

    def create_model_factory(self, model_name: str, service_name=None, index=None):
        for service in self.services:
            if model_name in service.available():
                if service_name is None or service_name == service._inference_service_:
                    return service.create_model(model_name)

        raise Exception(f"Model {model_name} not found in any of the services")



from sqlalchemy import Column, String
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


class Data(Base):
    __tablename__ = "data"
    key = Column(String, primary_key=True)
    value = Column(String)



import os
import aiohttp
import json
from typing import Any
from edsl.exceptions import MissingAPIKeyError
from edsl.language_models.LanguageModel import LanguageModel

from edsl.inference_services.InferenceServiceABC import InferenceServiceABC


class GoogleService(InferenceServiceABC):
    _inference_service_ = "google"

    @classmethod
    def available(cls):
        return ["gemini-pro"]

    @classmethod
    def create_model(
        cls, model_name: str = "gemini-pro", model_class_name=None
    ) -> LanguageModel:
        if model_class_name is None:
            model_class_name = cls.to_class_name(model_name)

        class LLM(LanguageModel):
            _model_ = model_name
            _inference_service_ = cls._inference_service_
            _parameters_ = {
                "temperature": 0.5,
                "topP": 1,
                "topK": 1,
                "maxOutputTokens": 2048,
                "stopSequences": [],
            }

            async def async_execute_model_call(
                self, user_prompt: str, system_prompt: str = ""
            ) -> dict[str, Any]:
                # self.api_token = os.getenv("GOOGLE_API_KEY")
                combined_prompt = user_prompt + system_prompt
                url = f"https://generativelanguage.googleapis.com/v1beta/models/{model_name}:generateContent?key={self.api_token}"
                headers = {"Content-Type": "application/json"}
                data = {
                    "contents": [{"parts": [{"text": combined_prompt}]}],
                    "generationConfig": {
                        "temperature": self.temperature,
                        "topK": self.topK,
                        "topP": self.topP,
                        "maxOutputTokens": self.maxOutputTokens,
                        "stopSequences": self.stopSequences,
                    },
                }

                async with aiohttp.ClientSession() as session:
                    async with session.post(
                        url, headers=headers, data=json.dumps(data)
                    ) as response:
                        raw_response_text = await response.text()
                        return json.loads(raw_response_text)

            def parse_response(self, raw_response: dict[str, Any]) -> str:
                data = raw_response
                return data["candidates"][0]["content"]["parts"][0]["text"]

        LLM.__name__ = model_name

        return LLM


if __name__ == "__main__":
    pass



"""Multiple choice question type."""
import textwrap
from edsl.prompts.QuestionInstructionsBase import QuestionInstuctionsBase


class MultipleChoiceTurbo(QuestionInstuctionsBase):
    """Multiple choice question type."""

    question_type = "multiple_choice"
    model = "gpt-3.5-turbo"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}
        Return a valid JSON formatted like this, selecting only the number of the option:
        {"answer": <put answer code here>, "comment": "<put explanation here>"}
        Only 1 option may be selected.
        """
    )


class MultipleChoice(QuestionInstuctionsBase):
    """Multiple choice question type."""

    question_type = "multiple_choice"
    model = "gpt-4-1106-preview"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        The options are
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}
        Return a valid JSON formatted like this, selecting only the number of the option:
        {"answer": <put answer code here>, "comment": "<put explanation here>"}
        Only 1 option may be selected.
        """
    )


class LikertFive(MultipleChoice):
    """Likert five question type."""

    question_type = "likert_five"


class YesNo(MultipleChoice):
    """Yes/No question type."""

    question_type = "yes_no"



from __future__ import annotations
import textwrap
from typing import Any, Optional
from edsl.questions.QuestionBase import QuestionBase
from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionFreeText(QuestionBase):
    """This question prompts the agent to respond with free text."""

    question_type = "free_text"
    default_instructions = textwrap.dedent(
        """\
        You are being asked the following question: {{question_text}}
        Return a valid JSON formatted like this: 
        {"answer": "<put free text answer here>"}
        """
    )

    def __init__(
        self,
        question_name: str,
        question_text: str,
        instructions: Optional[str] = None,
    ):
        """Instantiate a new QuestionFreeText.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionFreeText.default_instructions`.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.instructions = instructions

    ################
    # Answer methods
    ################
    def _validate_answer(self, answer: Any) -> dict[str, str]:
        """Validate the answer."""
        self._validate_answer_template_basic(answer)
        self._validate_answer_key_value(answer, "answer", str)
        return answer

    def _translate_answer_code_to_answer(self, answer, scenario: Scenario = None):
        """Do nothing, because the answer is already in a human-readable format."""
        return answer

    def _simulate_answer(self, human_readable: bool = True) -> dict[str, str]:
        """Simulate a valid answer for debugging purposes."""
        return {"answer": random_string()}

    @classmethod
    def example(cls) -> QuestionFreeText:
        """Return an example instance of a free text question."""
        return cls(question_name="how_are_you", question_text="How are you?")


def main():
    """Create an example question and demonstrate its functionality."""
    from edsl.questions.QuestionFreeText import QuestionFreeText

    q = QuestionFreeText.example()
    q.question_text
    q.question_name
    q.instructions
    # validate an answer
    q._validate_answer({"answer": "I like custard"})
    # translate answer code
    q._translate_answer_code_to_answer({"answer"})
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



"""This module contains the descriptors used to set the attributes of the Agent class."""
from typing import Dict
from edsl.utilities.utilities import is_valid_variable_name
from edsl.exceptions.agents import AgentNameError, AgentTraitKeyError


class NameDescriptor:
    """ABC for something."""

    def __get__(self, instance, owner):
        """Return the value of the attribute."""
        return instance.__dict__[self.name]

    def __set__(self, instance, name: str) -> None:
        """Set the value of the attribute."""
        instance.__dict__[self.name] = name

    def __set_name__(self, owner, name: str) -> None:
        """Set the name of the attribute."""
        self.name = "_" + name


class TraitsDescriptor:
    """ABC for something."""

    def __get__(self, instance, owner):
        """Return the value of the attribute."""
        return instance.__dict__[self.name]

    def __set__(self, instance, traits_dict: Dict[str, str]) -> None:
        """Set the value of the attribute."""
        for key, value in traits_dict.items():
            if key == "name":
                raise AgentNameError(
                    """Trait keys cannot be 'name'. Instead, use the 'name' attribute directly e.g., 
                    Agent(name="my_agent", traits={"trait1": "value1", "trait2": "value2"})
                    """
                )

            if not is_valid_variable_name(key):
                raise AgentTraitKeyError(
                    "Trait keys must be valid Python identifiers (must be alphanumeric, cannot start with a number and must use underscores instead of spaces)."
                )

        instance.__dict__[self.name] = traits_dict

    def __set_name__(self, owner, name: str) -> None:
        """Set the name of the attribute."""
        self.name = name


class CodebookDescriptor:
    """ABC for something."""

    def __get__(self, instance, owner):
        """Return the value of the attribute."""
        return instance.__dict__[self.name]

    def __set__(self, instance, codebook_dict: Dict[str, str]) -> None:
        """Set the value of the attribute."""
        instance.__dict__[self.name] = codebook_dict

    def __set_name__(self, owner, name: str) -> None:
        """Set the name of the attribute."""
        self.name = "_" + name


class InstructionDescriptor:
    """ABC for something."""

    def __get__(self, instance, owner):
        """Return the value of the attribute."""
        return instance.__dict__[self.name]

    def __set__(self, instance, instruction) -> None:
        """Set the value of the attribute."""
        instance.__dict__[self.name] = instruction
        instance.set_instructions = instruction != instance.default_instruction

    def __set_name__(self, owner, name: str) -> None:
        """Set the name of the attribute."""
        self.name = "_" + name



from __future__ import annotations
import random
import textwrap
from typing import Any, Optional, Union
from edsl.questions.QuestionBase import QuestionBase
from edsl.questions.descriptors import IntegerOrNoneDescriptor

from edsl.scenarios import Scenario
from edsl.utilities import random_string


class QuestionList(QuestionBase):
    """This question prompts the agent to answer by providing a list of items as comma-separated strings."""

    question_type = "list"
    max_list_items: int = IntegerOrNoneDescriptor()

    def __init__(
        self,
        question_name: str,
        question_text: str,
        max_list_items: Optional[int] = None,
    ):
        """Instantiate a new QuestionList.

        :param question_name: The name of the question.
        :param question_text: The text of the question.
        :param instructions: Instructions for the question. If not provided, the default instructions are used. To view them, run `QuestionList.default_instructions`.
        :param max_list_items: The maximum number of items that can be in the answer list.
        """
        self.question_name = question_name
        self.question_text = question_text
        self.max_list_items = max_list_items

    ################
    # Answer methods
    ################
    def _validate_answer(self, answer: Any) -> dict[str, Union[list[str], str]]:
        """Validate the answer."""
        self._validate_answer_template_basic(answer)
        self._validate_answer_key_value(answer, "answer", list)
        self._validate_answer_list(answer)
        return answer

    def _translate_answer_code_to_answer(self, answer, scenario: Scenario = None):
        """There is no answer code."""
        return answer

    def _simulate_answer(self, human_readable: bool = True):
        """Simulate a valid answer for debugging purposes (what the validator expects)."""
        num_items = random.randint(1, self.max_list_items or 2)
        return {"answer": [random_string() for _ in range(num_items)]}

    ################
    # Helpful methods
    ################
    @classmethod
    def example(cls) -> QuestionList:
        """Return an example of a list question."""
        return cls(
            question_name="list_of_foods",
            question_text="What are your favorite foods?",
            max_list_items=5,
        )


def main():
    """Create an example of a list question and demonstrate its functionality."""
    from edsl.questions.QuestionList import QuestionList

    q = QuestionList.example()
    q.question_text
    q.question_name
    q.max_list_items
    # validate an answer
    q._validate_answer({"answer": ["pasta", "garlic", "oil", "parmesan"]})
    # translate answer code
    q._translate_answer_code_to_answer(["pasta", "garlic", "oil", "parmesan"])
    # simulate answer
    q._simulate_answer()
    q._simulate_answer(human_readable=False)
    q._validate_answer(q._simulate_answer(human_readable=False))
    # serialization (inherits from Question)
    q.to_dict()
    assert q.from_dict(q.to_dict()) == q

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from abc import abstractmethod, ABC
from typing import Any
import re


class InferenceServiceABC(ABC):
    """Abstract class for inference services."""

    @abstractmethod
    def available() -> list[str]:
        pass

    @abstractmethod
    def create_model():
        pass

    @staticmethod
    def to_class_name(s):
        s = re.sub(r"[^a-zA-Z0-9 ]", "", s)
        s = "".join(word.title() for word in s.split())
        if s and s[0].isdigit():
            s = "Class" + s
        return s


if __name__ == "__main__":
    pass
    # deep_infra_service = DeepInfraService("deep_infra", "DEEP_INFRA_API_KEY")
    # deep_infra_service.available()
    # m = deep_infra_service.create_model("microsoft/WizardLM-2-7B")
    # response = m().hello()
    # print(response)

    # anthropic_service = AnthropicService("anthropic", "ANTHROPIC_API_KEY")
    # anthropic_service.available()
    # m = anthropic_service.create_model("claude-3-opus-20240229")
    # response = m().hello()
    # print(response)
    # factory = OpenAIService("openai", "OPENAI_API")
    # factory.available()
    # m = factory.create_model("gpt-3.5-turbo")
    # response = m().hello()

    # from edsl import QuestionFreeText
    # results = QuestionFreeText.example().by(m()).run()

    # collection = InferenceServicesCollection([
    #     OpenAIService,
    #     AnthropicService,
    #     DeepInfraService
    # ])

    # available = collection.available()
    # factory = collection.create_model_factory(*available[0])
    # m = factory()
    # from edsl import QuestionFreeText
    # results = QuestionFreeText.example().by(m).run()
    # print(results)



from collections import UserList


class TasksList(UserList):
    pass
    # def status(self, debug=False):
    #     if debug:
    #         for task in self:
    #             print(f"Task {task.edsl_name}")
    #             print(f"\t DEPENDS ON: {task.depends_on}")
    #             print(f"\t DONE: {task.done()}")
    #             print(f"\t CANCELLED: {task.cancelled()}")
    #             if not task.cancelled():
    #                 if task.done():
    #                     print(f"\t RESULT: {task.result()}")
    #                 else:
    #                     print(f"\t RESULT: None - Not done yet")

    #         print("---------------------")



"""This module contains the data transfer models for the application."""
from collections import UserDict


class AgentResponseDict(UserDict):
    """A dictionary to store the response of the agent to a question."""

    def __init__(
        self,
        *,
        question_name,
        answer,
        prompts,
        usage=None,
        comment=None,
        cached_response=None,
        raw_model_response=None,
        simple_model_raw_response=None,
    ):
        """Initialize the AgentResponseDict object."""
        usage = usage or {"prompt_tokens": 0, "completion_tokens": 0}
        super().__init__(
            {
                "answer": answer,
                "comment": comment,
                "question_name": question_name,
                "prompts": prompts,
                "usage": usage,
                "cached_response": cached_response,
                "raw_model_response": raw_model_response,
                "simple_model_raw_response": simple_model_raw_response,
            }
        )



"""A module to represent a dataset of observations."""
from __future__ import annotations
import numpy as np
from collections import UserList
from typing import Any
from edsl.results.ResultsExportMixin import ResultsExportMixin


class Dataset(UserList, ResultsExportMixin):
    """A class to represent a dataset of observations."""

    def __init__(self, data: list[dict[str, Any]] = None):
        """Initialize the dataset with the given data."""
        super().__init__(data)

    def relevant_columns(self) -> set:
        """Return the set of keys that are present in the dataset."""
        return set([list(result.keys())[0] for result in self.data])

    def _key_to_value(self, key: str) -> Any:
        """Retrieve the value associated with the given key from the dataset."""
        for d in self.data:
            if key in d:
                return d[key]
        else:
            raise KeyError(f"Key '{key}' not found in any of the dictionaries.")

    def first(self) -> dict[str, Any]:
        """Get the first value of the first key in the first dictionary."""

        def get_values(d):
            """Get the values of the first key in the dictionary."""
            return list(d.values())[0]

        return get_values(self.data[0])[0]

    def _repr_html_(self) -> str:
        """Return an HTML representation of the dataset."""
        from edsl.utilities.utilities import data_to_html

        return data_to_html(self.data)

    def order_by(self, sort_key: str, reverse: bool = False) -> Dataset:
        """Return a new dataset with the observations sorted by the given key."""

        def sort_indices(lst: list[Any]) -> list[int]:
            """
            Return the indices that would sort the list.

            :param lst: The list to be sorted.
            :return: A list of indices that would sort the list.
            """
            indices = np.argsort(lst).tolist()
            if reverse:
                indices.reverse()
            return indices

        if not any(sort_key in d for d in self.data):
            raise ValueError(f"Key '{sort_key}' not found in any of the dictionaries.")

        relevant_values = self._key_to_value(sort_key)
        sort_indices_list = sort_indices(relevant_values)
        new_data = []
        for observation in self.data:
            print(observation)
            key, values = list(observation.items())[0]
            new_values = [values[i] for i in sort_indices_list]
            new_data.append({key: new_values})

        return Dataset(new_data)



from collections import UserDict


class TokensUsed(UserDict):
    """ "Container for tokens used by a task."""

    def __init__(self, cached_tokens, new_tokens):
        d = {"cached_tokens": cached_tokens, "new_tokens": new_tokens}
        super().__init__(d)


if __name__ == "__main__":
    pass



from typing import Callable, Union, List
from collections import UserDict

from edsl.jobs.tokens.TokenUsage import TokenUsage
from edsl.jobs.interviews.InterviewStatusDictionary import InterviewStatusDictionary
from edsl.jobs.tokens.InterviewTokenUsage import InterviewTokenUsage


class TaskCreators(UserDict):
    """A dictionary of task creators. A task is one question being answered.

    This is used to track the status of the tasks within an interview.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def token_usage(self) -> InterviewTokenUsage:
        """Determines how many tokens were used for the interview."""
        cached_tokens = TokenUsage(from_cache=True)
        new_tokens = TokenUsage(from_cache=False)
        for task_creator in self.values():
            token_usage = task_creator.token_usage()
            cached_tokens += token_usage["cached_tokens"]
            new_tokens += token_usage["new_tokens"]
        return InterviewTokenUsage(
            new_token_usage=new_tokens, cached_token_usage=cached_tokens
        )

    def print(self):
        from rich import print

        print({task.get_name(): task.task_status for task in self.values()})

    @property
    def interview_status(self) -> InterviewStatusDictionary:
        """Returns a dictionary, InterviewStatusDictionary, mapping task status codes to counts of tasks in that state."""
        status_dict = InterviewStatusDictionary()
        for task_creator in self.values():
            status_dict[task_creator.task_status] += 1
            status_dict["number_from_cache"] += task_creator.from_cache
        return status_dict

    def status_logs(self):
        """Returns a list of status logs for each task."""
        return [task_creator.status_log for task_creator in self.values()]
